UNIT MCB;

{$mode objfpc}{$H+}

INTERFACE

USES
  CONTNRS;

CONST
  wordsize   = 4;
  idlen      = 32;
  bufsize    = 256;

  maxexp     = 308;
  minexp     = -324;
  maxreal    = 3.4E+38;
  minreal    = -1.5E-45;
  maxlreal   = 1.7E+308;
  minlreal   = -5.0E-324;
  maxint     = 2147483647;
  minint     = -2147483647-1;
  maxhuge    = 9223372036854775807;
  minhuge    = -9223372036854775807-1;

  cHead   = 0; cConst = 1;  cVars    = 2;  cPar    = 3; cFld  = 4;
  cTyp    = 5; cSProc = 6;  cSFunc   = 7;  cTProc  = 8; cMod  = 9;
  cNode   = 10;
  fByte   = 1; fBool  = 2;  fChar    = 3;  fInt    = 4;
  fReal   = 5; fSet   = 6;  fPointer = 7;  fNilTyp = 8;
  fNoTyp  = 9; fProc  = 10; fString  = 11; fArray  = 12;
  fRecord = 13;fLreal = 14; fHuge    = 15;

TYPE
  Ident   = ARRAY [0..idlen-1]   OF CHAR;
  Strings = ARRAY [0..bufsize-1] OF CHAR;

  Object_ = CLASS;
  Type_   = CLASS;

  Item = CLASS
    a,b    : Item;
    typ    : Type_;
    op     : INTEGER;
    reg    : INTEGER;
    clas   : INTEGER;
    rdo    : BOOLEAN;
    neg    : BOOLEAN;
    val    : EXTENDED;
  END;

  Object_ = CLASS(Item)
    name   : Ident;
    next   : Object_;
    dsc    : Object_;
    lev    : INTEGER;
    exno   : INTEGER;
    expo   : BOOLEAN;
  END;

  Type_ = CLASS
    base   : Type_;
    dsc    : Object_;
    typobj : Object_;
    form   : INTEGER;
    ref    : INTEGER;
    mno    : INTEGER;
    nofpar : INTEGER;
    len    : INTEGER;
    size   : INTEGER;
  END;

VAR
  GC       : TObjectList;
  Mark     : PROCEDURE (msg: ARRAY OF CHAR);
  syscount : INTEGER;

  version  : INTEGER;
  errpos   : INTEGER;
  errcount : INTEGER;
  prevcol  : INTEGER;
  prevline : INTEGER;

  topScope,  _system,  universe, dummy             : Object_;
  byteType,  boolType, charType, intType, hugeType : Type_;
  lrealType, realType, setType,  nilType, noType   : Type_;
  strType                                          : Type_;

  PROCEDURE NEW(VAR x: Type_);
  PROCEDURE NEW(VAR x: Item);    OVERLOAD;
  PROCEDURE NEW(VAR x: Object_); OVERLOAD;
  PROCEDURE Clear;

  FUNCTION  NewItem(clas: INTEGER; a,b: Item; op: INTEGER): Item;
  PROCEDURE MakeConstItem(VAR x: Item; typ: Type_; val: EXTENDED);

IMPLEMENTATION

PROCEDURE NEW(VAR x: Item);    BEGIN x := Item.Create; GC.Add(x);    END;
PROCEDURE NEW(VAR x: Object_); BEGIN x := Object_.Create; GC.Add(x); END;
PROCEDURE NEW(VAR x: Type_);   BEGIN x := Type_.Create; GC.Add(x);   END;

PROCEDURE MarkProc(msg: ARRAY OF CHAR);
BEGIN
  IF errcount = 0 THEN BEGIN
    errpos := prevcol;
  END ELSE IF (errpos <> prevcol) AND (errcount < 25) THEN BEGIN
  END;
  INC(errcount);
  errpos := prevcol;
END;

FUNCTION NewItem(clas: INTEGER; a,b: Item; op: INTEGER): Item;
VAR
  x: Item;
BEGIN
  NEW(x);
  x.clas := clas;
  x.typ  := noType;
  x.op   := op;
  x.rdo  := FALSE;
  x.neg  := FALSE;
  x.val  := 0;
  x.reg  := 1;
  x.a    := a;
  x.b    := b;
  RESULT := x;
END;

PROCEDURE MakeConstItem(VAR x: Item; typ: Type_; val: EXTENDED);
BEGIN
  NEW(x);
  x.clas := cConst;
  x.typ  := typ;
  x.val  := val;
  x.reg  := -1;
  x.op   := 0;
  x.rdo  := FALSE;
  x.neg  := FALSE;
  x.a    := NIL;
  x.b    := NIL;
END;

PROCEDURE OpenScope;
VAR
  scope: Object_;
BEGIN
  NEW(scope);
  scope.clas := cHead;
  scope.dsc  := topScope;
  scope.next := NIL;
  topScope   := scope;
END;

PROCEDURE CloseScope;
BEGIN
  topScope := topScope.dsc
END;

PROCEDURE Enter(name: ARRAY OF CHAR; clas: INTEGER; typ: Type_; val: EXTENDED);
VAR
  obj: Object_;
BEGIN
  NEW(obj);
  obj.name := COPY(name,0,LENGTH(name));
  obj.clas := clas;
  obj.typ  := typ;
  obj.val  := val;
  obj.dsc  := NIL;
  IF clas  = cTyp THEN typ.typobj := obj;
  obj.next := _system;
  _system  := obj;
END;

PROCEDURE Create(VAR typ: Type_; ref, form: INTEGER; size: LONGINT);
BEGIN
  NEW(typ);
  typ.form := form;
  typ.size := size;
  typ.ref  := ref;
  typ.base := NIL;
END;

PROCEDURE Clear;
BEGIN
  errpos   := 0;
  errcount := 0;
  dummy.free;
  WHILE GC.Count > syscount DO BEGIN
    GC.Items[GC.Count-1].free;
    GC.Delete(GC.Count-1);
  END;
  dummy      := Object_.Create;
  dummy.clas := cVars;
  dummy.typ  := intType;
END;

INITIALIZATION
  Mark     := @MarkProc;
  GC       := TObjectList.Create(false);

  syscount := 0;
  topScope := universe;
  _system  := NIL;

  Create(byteType ,fByte  ,fInt    , 1);
  Create(boolType ,fBool  ,fBool   , 1);
  Create(charType ,fChar  ,fChar   , 1);
  Create(intType  ,fInt   ,fInt    , 4);
  Create(realType ,fReal  ,fReal   , 4);
  Create(setType  ,fSet   ,fSet    , 4);
  Create(nilType  ,fNilTyp,fNilTyp , 4);
  Create(noType   ,fNoTyp ,fNoTyp  , 4);
  Create(strType  ,fString,fString , 8);
  Create(hugeType ,fHuge  ,fHuge   , 8);
  Create(lrealType,fLReal ,fLReal  , 8);

  Enter('SET'     , cTyp, setType  , 0);
  Enter('BYTE'    , cTyp, byteType , 0);
  Enter('CHAR'    , cTyp, charType , 0);
  Enter('REAL'    , cTyp, realType , 0);
  Enter('BOOLEAN' , cTyp, boolType , 0);
  Enter('INTEGER' , cTyp, intType  , 0);
  Enter('HUGEINT' , cTyp, hugeType , 0);
  Enter('LONGREAL', cTyp, lrealType, 0);

  topScope := NIL;
  OpenScope;
  topScope.next := _system;
  universe   := topScope;
  _system    := NIL;
  syscount   := GC.Count;

  dummy      := Object_.Create;
  dummy.clas := cVars;
  dummy.typ  := intType;

FINALIZATION
  WHILE GC.Count > 0 DO BEGIN
    GC.Items[GC.Count-1].free;
    GC.Delete(GC.Count-1);
  END;
  GC.free;
  dummy.free;
END.

