UNIT MCE;

{$mode objfpc}{$H+}

INTERFACE

USES
  MCB, MCS;

VAR
  sym  : INTEGER;
  dc   : INTEGER;
  level: INTEGER;
  exno : INTEGER;

  PROCEDURE Expression(VAR x: MCB.Item);

IMPLEMENTATION

PROCEDURE Check(s: INTEGER; msg: ARRAY OF CHAR);
BEGIN
  IF sym = s THEN MCS.Get(sym) ELSE MCB.Mark(msg);
END;

PROCEDURE CheckType(op: INTEGER; VAR x,y: MCB.Item);
VAR
  xf : INTEGER;
  yf : INTEGER;
BEGIN

  xf := x.typ.form;

  CASE op OF
    MCS.lx_times,
    MCS.lx_plus,
    MCS.lx_minus  :
                    IF xf IN [MCB.fInt,MCB.fHuge,MCB.fReal,MCB.fLreal,MCB.fSet] THEN
                      yf := xf
                    ELSE BEGIN
                      MCB.Mark('bad type 1');
                      x.typ := MCB.hugeType;
                    END;
    MCS.lx_rdiv   :
                    IF xf IN [MCB.fReal,MCB.fLreal,MCB.fSet] THEN
                      yf := xf
                    ELSE BEGIN
                      MCB.Mark('bad type 2');
                      x.typ := MCB.lrealType;
                    END;
    MCS.lx_div,
    MCS.lx_mod    :
                    IF xf IN [MCB.fInt,MCB.fHuge] THEN
                      yf := xf
                    ELSE BEGIN
                      MCB.Mark('bad type 3');
                      x.typ := MCB.hugeType;
                    END;
    MCS.lx_and,
    MCS.lx_or     :
                    IF xf IN [MCB.fBool] THEN
                      yf := xf
                    ELSE BEGIN
                      MCB.Mark('bad type 4');
                      x.typ := MCB.boolType;
                    END;
  END;

  IF y.typ.form <> yf THEN BEGIN
    CASE yf OF
      MCB.fBool  : BEGIN MCB.Mark('not Boolean');  y.typ := MCB.boolType; END;
      MCB.fInt   : BEGIN MCB.Mark('not Integer');  y.typ := MCB.intType;  END;
      MCB.fHuge  : BEGIN MCB.Mark('not Hugeint');  y.typ := MCB.hugeType; END;
      MCB.fReal  : BEGIN MCB.Mark('not Real');     y.typ := MCB.realType; END;
      MCB.fLreal : BEGIN MCB.Mark('not Longreal'); y.typ := MCB.realType; END;
      MCB.fSet   : BEGIN MCB.Mark('not Set');      y.typ := MCB.setType;  END;
    END;
  END;

  x := MCB.NewItem(MCB.cNode,x,y,op);
  x.typ := y.typ;

END;

PROCEDURE Factor(VAR x: MCB.Item);
VAR
  obj: MCB.Object_;
BEGIN

  IF (sym < MCS.lx_char) OR (sym > MCS.lx_ident) THEN BEGIN
    MCB.Mark('expression expected');
    REPEAT
      MCS.Get(sym)
    UNTIL (sym >= MCS.lx_char) AND (sym <= MCS.lx_ident);
  END;

  IF sym = MCS.lx_ident THEN BEGIN

  END ELSE IF sym = MCS.lx_int    THEN BEGIN MCB.MakeConstItem(x, MCB.intType,   MCS.r64);  MCS.Get(sym);
  END ELSE IF sym = MCS.lx_real   THEN BEGIN MCB.MakeConstItem(x, MCB.realType,  MCS.r64);  MCS.Get(sym);
  END ELSE IF sym = MCS.lx_huge   THEN BEGIN MCB.MakeConstItem(x, MCB.hugeType,  MCS.r64);  MCS.Get(sym);
  END ELSE IF sym = MCS.lx_lreal  THEN BEGIN MCB.MakeConstItem(x, MCB.lrealType, MCS.r64);  MCS.Get(sym);
  END ELSE IF sym = MCS.lx_char   THEN BEGIN MCB.MakeConstItem(x, MCB.charType,  MCS.r64);  MCS.Get(sym);
  END ELSE IF sym = MCS.lx_nil    THEN BEGIN MCB.MakeConstItem(x, MCB.nilType,  0);         MCS.Get(sym);
  END ELSE IF sym = MCS.lx_false  THEN BEGIN MCB.MakeConstItem(x, MCB.boolType, 0);         MCS.Get(sym);
  END ELSE IF sym = MCS.lx_true   THEN BEGIN MCB.MakeConstItem(x, MCB.boolType, 1);         MCS.Get(sym);
  //END ELSE IF sym = MCS.lx_string THEN BEGIN MCB.MakeStringItem(x, MCS.slen, MCS.strs);     MCS.Get(sym);
  END ELSE IF sym = MCS.lx_lparen THEN BEGIN MCS.Get(sym); Expression(x); Check(MCS.lx_rparen, 'no )')
  //END ELSE IF sym = MCS.lx_lbrace THEN BEGIN MCS.Get(sym); Sets(x);       Check(MCS.lx_rbrace, 'no }')
  //END ELSE IF sym = MCS.lx_not    THEN BEGIN MCS.Get(sym); Factor(x);     CheckType(x,MCB.fBool); //OCG.Not1(x)
  END ELSE BEGIN
    MCB.Mark('not a factor');
    MCB.MakeConstItem(x, MCB.intType, 0);
  END;

END;

PROCEDURE Term(VAR x: MCB.Item);
VAR
  y: MCB.Item;
  op, f: INTEGER;
BEGIN
  Factor(x);

  IF x.typ = NIL THEN x.typ := MCB.noType;
  f := x.typ.form;

  WHILE (sym >= MCS.lx_times) AND (sym <= MCS.lx_and) DO BEGIN
    op := sym;
    MCS.Get(sym);
    Factor(y);
    CheckType(op,x,y);
    (*
    IF op = MCS.lx_times THEN BEGIN
      IF          f = MCB.fInt  THEN BEGIN Factor(y); CheckType(op,x,y,MCB.fInt);  //OCG.MulOp(x, y)
      END ELSE IF f = MCB.fReal THEN BEGIN Factor(y); CheckType(op,x,y,MCB.fReal); //OCG.RealOp(op, x, y)
      END ELSE IF f = MCB.fSet  THEN BEGIN Factor(y); CheckType(op,x,y,MCB.fSet);  //OCG.SetOp(op, x, y)
      END ELSE
        MCB.Mark('bad type');
    END ELSE IF (op = MCS.lx_div) OR (op = MCS.lx_mod) THEN BEGIN
      Factor(y);
      CheckType(op,x,y,MCB.fInt);
      //OCG.DivOp(op, x, y)
    END ELSE IF op = MCS.lx_rdiv THEN BEGIN
      IF          f = MCB.fReal THEN BEGIN Factor(y); CheckType(op,x,y,MCB.fReal); //OCG.RealOp(op, x, y)
      END ELSE IF f = MCB.fSet  THEN BEGIN Factor(y); CheckType(op,x,y,MCB.fSet);  //OCG.SetOp(op, x, y)
      END ELSE
        MCB.Mark('bad type');
    END ELSE BEGIN
      //OCG.And1(x);
      Factor(y);
      CheckType(op,x,y,MCB.fBool);
      //OCG.And2(x, y)
    END;
    *)
  END;
END;

PROCEDURE SimpleExpression(VAR x: MCB.Item);
VAR
  y: MCB.Item;
  op: INTEGER;
BEGIN

  y := NIL;

  IF sym = MCS.lx_minus THEN BEGIN
    MCS.Get(sym);
    Term(x);
    IF x.typ.form IN [MCB.fInt, MCB.fHuge, MCB.fReal, MCB.fLreal, MCB.fSet] THEN x.neg:= NOT x.neg ELSE CheckType(MCS.lx_minus,x,y);//CheckType(MCS.lx_minus,x,y,MCB.fInt);
  END ELSE IF sym = MCS.lx_plus THEN BEGIN
    MCS.Get(sym);
    Term(x);
  END ELSE
    Term(x);

  WHILE (sym >= MCS.lx_plus) AND (sym <= MCS.lx_or) DO BEGIN
    op := sym;
    MCS.Get(sym);
    Term(y);
    CheckType(op,x,y);
    (*
    IF op = MCS.lx_or THEN BEGIN
      //OCG.Or1(x);
      Term(y);
      CheckType(op,x,y,MCB.fBool);
      //OCG.Or2(x, y)
    END ELSE IF x.typ.form = MCB.fInt THEN BEGIN
      Term(y);
      CheckType(op,x,y,MCB.fInt);
      //OCG.AddOp(op, x, y)
    END ELSE IF x.typ.form = MCB.fReal THEN BEGIN
      Term(y);
      CheckType(op,x,y,MCB.fReal);
      //OCG.RealOp(op, x, y)
    END ELSE BEGIN
      Term(y);
      CheckType(op,x,y,MCB.fSet);
      //OCG.SetOp(op, x, y)
    END;
    *)
  END;

END;

PROCEDURE Expression(VAR x: MCB.Item);
VAR
  y: MCB.Item;
  obj: MCB.Object_;
  rel, xf, yf: INTEGER;
BEGIN

  SimpleExpression(x);

  IF (sym >= MCS.lx_eql) AND (sym <= MCS.lx_geq) THEN BEGIN

    rel := sym;
    MCS.Get(sym);
    SimpleExpression(y);
    xf := x.typ.form;
    yf := y.typ.form;

    IF x.typ = y.typ THEN BEGIN

      IF (xf IN [MCB.fChar, MCB.fInt]) THEN
        //OCG.IntRelation(rel, x, y)
      ELSE IF xf = MCB.fReal THEN
        //OCG.RealRelation(rel, x, y)
      ELSE IF (xf IN [MCB.fSet, MCB.fPointer, MCB.fProc, MCB.fNilTyp, MCB.fBool]) THEN BEGIN
        IF rel <= MCS.lx_neq THEN
          //OCG.IntRelation(rel, x, y)
        ELSE
          MCB.Mark('only = or #');
      END ELSE IF (xf = MCB.fArray) AND (x.typ.base.form = MCB.fChar) OR (xf = MCB.fString) THEN
        //OCG.StringRelation(rel, x, y)
      ELSE
        MCB.Mark('illegal comparison');

    END ELSE IF (xf IN [MCB.fPointer, MCB.fProc]) AND (yf = MCB.fNilTyp)
    OR (yf IN [MCB.fPointer, MCB.fProc]) AND (xf = MCB.fNilTyp) THEN BEGIN

      IF rel <= MCS.lx_neq THEN
        //OCG.IntRelation(rel, x,  y)
      ELSE
        MCB.Mark('only = or #');

    END ELSE IF (xf = MCB.fPointer) AND (yf = MCB.fPointer)
    //AND (IsExtension(x.typ.base, y.typ.base) OR IsExtension(y.typ.base, x.typ.base))
    THEN BEGIN

      IF rel <= MCS.lx_neq THEN
        //OCG.IntRelation(rel,  x, y)
      ELSE
        MCB.Mark('only = or #');

    END ELSE IF (xf = MCB.fArray) AND (x.typ.base.form = MCB.fChar) AND
    ((yf = MCB.fString) OR (yf = MCB.fArray) AND (y.typ.base.form = MCB.fChar))
    OR (yf = MCB.fArray) AND (y.typ.base.form = MCB.fChar) AND (xf = MCB.fString) THEN
      //OCG.StringRelation(rel, x, y)
    ELSE IF (xf = MCB.fChar) AND (yf = MCB.fString) AND (y.b.val = 2) THEN BEGIN
      //MCB.StrToChar(y);
      //OCG.IntRelation(rel, x, y)
    END ELSE IF (yf = MCB.fChar) AND (xf = MCB.fString) AND (x.b.val = 2) THEN BEGIN
      //MCB.StrToChar(x);
      //OCG.IntRelation(rel, x, y)
    END ELSE IF (xf = MCB.fInt) AND (yf = MCB.fInt) THEN
      //OCG.IntRelation(rel,  x, y)
    ELSE
      MCB.Mark('illegal comparison');
    x.typ := MCB.boolType;

  END ELSE IF sym = MCS.lx_in THEN BEGIN

    MCS.Get(sym);
    //CheckType(x,MCB.fInt);
    SimpleExpression(y);
    //CheckType(y,MCB.fSet);
    //OCG.In2(x, y) ;
    x.typ := MCB.boolType;

  END ELSE IF sym = MCS.lx_is THEN BEGIN

    MCS.Get(sym);
    //Qualident(obj);
    //TypeTest(x, obj.typ, FALSE) ;
    x.typ := MCB.boolType;

  END;

END;

END.

