UNIT MCS;

{$mode objfpc}{$H+}
(* ANSI   ->  -> ANSI *)

INTERFACE

USES
  RTL, MCB;

CONST
  lx_null       = 0;  lx_times  = 1;  lx_rdiv      = 2;  lx_div       = 3;
  lx_mod        = 4;  lx_and    = 5;  lx_plus      = 6;  lx_minus     = 7;
  lx_or         = 8;  lx_eql    = 9;  lx_neq       = 10; lx_lss       = 11;
  lx_leq        = 12; lx_gtr    = 13; lx_geq       = 14; lx_in        = 15;
  lx_is         = 16; lx_arrow  = 17; lx_period    = 18; lx_char      = 20;
  lx_int        = 21; lx_huge   = 22; lx_real      = 23; lx_lreal     = 24;
  lx_false      = 25; lx_true   = 26; lx_nil       = 27; lx_string    = 28;
  lx_not        = 29; lx_lparen = 30; lx_lbrak     = 31; lx_lbrace    = 32;
  lx_ident      = 33; lx_if     = 34; lx_while     = 35; lx_repeat    = 36;
  lx_case       = 37; lx_for    = 38; lx_comma     = 40; lx_colon     = 41;
  lx_becomes    = 42; lx_upto   = 43; lx_rparen    = 44; lx_rbrak     = 45;
  lx_rbrace     = 46; lx_then   = 47; lx_of        = 48; lx_do        = 49;
  lx_to         = 50; lx_by     = 51; lx_semicolon = 52; lx_end       = 53;
  lx_bar        = 54; lx_else   = 55; lx_elsif     = 56; lx_until     = 57;
  lx_return     = 58; lx_array  = 60; lx_record    = 61; lx_pointer   = 62;
  lx_const      = 63; lx_type   = 64; lx_var       = 65; lx_procedure = 66;
  lx_begin      = 67; lx_import = 68; lx_module    = 69; lx_eot       = 70;
  lx_definition = 71; lx_utf    = 72;

VAR
  id  : MCB.Ident;
  r64 : EXTENDED;
  slen: INTEGER;
  strs: MCB.Strings;
  EOF : BOOLEAN;

  PROCEDURE Get(VAR sym: INTEGER);
  PROCEDURE Init(addr: POINTER);

  PROCEDURE Push(sym: INTEGER);
  PROCEDURE Pop(VAR sym: INTEGER);
  FUNCTION  Find(name: MCB.Ident): BOOLEAN;
  PROCEDURE Add(name: MCB.Ident);
  PROCEDURE Clear;

IMPLEMENTATION

TYPE
  List = CLASS
    name       : MCB.Ident;
    next       : List;
  END;

  Data = CLASS
    ch         : CHAR;
    id         : MCB.Ident;
    eof        : BOOLEAN;
    r64        : EXTENDED;
    slen       : INTEGER;
    strs       : MCB.Strings;
    sym        : INTEGER;
    src        : POINTER;
    col        : INTEGER;
    line       : INTEGER;
    next       : Data;
  END;

CONST
  NKW        = 34;
  maxKWX     = 11;
  maxKWD     = 11;

VAR
  ch         : CHAR;
  src        : POINTER;
  col        : INTEGER;
  line       : INTEGER;
  scandata   : Data;
  namelist   : List;

  k        : INTEGER;
  KWX      : ARRAY [0..maxKWX-1] OF INTEGER;
  keyTab   : ARRAY [0..NKW-1] OF RECORD
    sym    : INTEGER;
    id     : ARRAY [0..maxKWD-1] OF CHAR;
  END;

PROCEDURE Push(sym: INTEGER);
VAR
  x: Data;
BEGIN
  x          := Data.Create;
  x.ch       := ch;
  x.id       := id;
  x.eof      := eof;
  x.sym      := sym;
  x.r64      := r64;
  x.slen     := slen;
  x.strs     := strs;
  x.col      := col;
  x.line     := line;
  x.src      := src;
  x.next     := scandata;
  scandata   := x;
END;

PROCEDURE Pop(VAR sym: INTEGER);
VAR
  x: Data;
BEGIN
  IF scandata <> NIL THEN BEGIN
    ch       := scandata.ch;
    id       := scandata.id;
    eof      := scandata.eof;
    sym      := scandata.sym;
    r64      := scandata.r64;
    slen     := scandata.slen;
    strs     := scandata.strs;
    col      := scandata.col;
    line     := scandata.line;
    src      := scandata.src;
    x        := scandata;
    scandata := x.next;
    x.free;
  END
END;

PROCEDURE Clear;

  PROCEDURE DataClear(VAR x: Data);
  BEGIN
    IF x=NIL THEN EXIT;
    IF x.next<>NIL THEN DataClear(x.next);
    x.free;
    x := NIL;
  END;

  PROCEDURE ListClear(VAR x: List);
  BEGIN
    IF x=NIL THEN EXIT;
    IF x.next<>NIL THEN ListClear(x.next);
    x.free;
    x := NIL;
  END;

BEGIN
  IF scandata <> NIL THEN DataClear(scandata);
  IF namelist <> NIL THEN ListClear(namelist);
END;

PROCEDURE Add(name: MCB.Ident);
VAR
  x: List;
BEGIN
  IF namelist <> NIL THEN BEGIN
    x := List.Create;
    x.name := name;
    x.next := namelist;
    namelist := x;
  END ELSE BEGIN
    namelist := List.Create;
    namelist.name := name;
    scandata.next := NIL;
  END;
END;

FUNCTION Find(name: MCB.Ident): BOOLEAN;
VAR
  x: List;
BEGIN
  RESULT := name = '';
  IF namelist <> NIL THEN BEGIN
    x := namelist;
    WHILE (x<>NIL) AND NOT RESULT DO BEGIN
      RESULT := x.name=name;
      x := x.next;
    END;
  END;
END;

PROCEDURE Read(VAR ch: CHAR);
BEGIN
  ch := CHAR(src^);
  IF ch <> #0 THEN BEGIN
    IF ch = #$0A THEN BEGIN
      col := 0;
      INC(line);
    END ELSE
      INC(col);
    INC(src);
  END ELSE BEGIN
    EOF := TRUE;
    ch  := #0;
  END;
END;

PROCEDURE Init(addr: POINTER);
BEGIN
  src          := addr;
  col          := 0;
  line         := 1;
  MCB.errpos   := 0;
  MCB.errcount := 0;
  MCB.prevcol  := col;
  MCB.prevline := line;
  EOF          := FALSE;
  Read(ch);
END;

PROCEDURE Identifier(VAR sym: INTEGER);
VAR
  i, k: INTEGER;
BEGIN
  i := 0;
  REPEAT
    IF i < MCB.idlen-1 THEN BEGIN id[i] := ch; INC(i) END ;
    Read(ch);
  UNTIL ((ch < '0') OR (ch > '9') AND (ch < 'A') OR (ch > 'Z') AND (ch < 'a') OR (ch > 'z') AND (ch <> '_') OR (i >= MCB.idlen));
  id[i] := #0;
  IF i < maxKWX THEN BEGIN
    k := KWX[i-1];
    WHILE (id <> keyTab[k].id) AND (k < KWX[i]) DO INC(k);
    IF k < KWX[i] THEN sym := keyTab[k].sym ELSE sym := lx_ident;
  END ELSE
    sym := lx_ident;
END;

PROCEDURE TextString;
VAR
  i: INTEGER;
BEGIN
  i := 0;
  Read(ch);
  WHILE NOT EOF AND (ch <> '"') DO BEGIN
    IF ch >= ' ' THEN BEGIN
      IF i < MCB.bufsize-1 THEN BEGIN
        strs[i] := ch;
        INC(i)
      END ELSE
        MCB.Mark('string too long');
    END;
    Read(ch);
  END ;
  strs[i] := #0;
  INC(i);
  Read(ch);
  slen := i;
END;

PROCEDURE HexString;
VAR
  i, m, n: INTEGER;
BEGIN
  i := 0;
  Read(ch);
  WHILE NOT EOF AND (ch <> '$') DO BEGIN
    WHILE (ch = ' ') OR (ch = #9) OR (ch = #$0D) OR (ch = #$0A) DO Read(ch);
    IF      ('0' <= ch) AND (ch <= '9') THEN m := ORD(ch) - $30
    ELSE IF ('A' <= ch) AND (ch <= 'F') THEN m := ORD(ch) - $37
    ELSE BEGIN
      m := 0;
      MCB.Mark('hexdig expected');
    END;
    Read(ch);
    IF      ('0' <= ch) AND (ch <= '9') THEN n := ORD(ch) - $30
    ELSE IF ('A' <= ch) AND (ch <= 'F') THEN n := ORD(ch) - $37
    ELSE BEGIN
      n := 0;
      MCB.Mark('hexdig expected');
    END;
    IF i < MCB.bufsize THEN BEGIN
      strs[i] := CHR(m*$10 + n);
      INC(i);
    END ELSE
      MCB.Mark('string too long');
    Read(ch);
  END;
  Read(ch);
  slen := i;
END;

PROCEDURE Comment;
BEGIN
  Read(ch);
  REPEAT
    WHILE NOT EOF AND (ch <> '*') DO BEGIN
      IF ch = '(' THEN BEGIN
        Read(ch);
        IF ch = '*' THEN Comment;
      END ELSE
        Read(ch);
    END;
    WHILE ch = '*' DO Read(ch)
  UNTIL (ch = ')') OR EOF;
  IF NOT EOF THEN
    Read(ch)
  ELSE
    MCB.Mark('unterminated comment');
END;

FUNCTION Ten(e: INT64): EXTENDED;
VAR
  x, t: EXTENDED;
BEGIN
  x := 1.0;
  t := 10.0;
  WHILE e > 0 DO BEGIN
    IF ODD(e) THEN x := t * x;
    t := t * t;
    e := RTL._DIV(e,2);
  END ;
  RESULT := x;
END;

PROCEDURE Number(VAR sym: INTEGER);
VAR
  i,n,s,h: INTEGER;
  e: INT64;
  x,k: EXTENDED;
  d: ARRAY [0..21-1] OF INTEGER;
  negE: BOOLEAN;
BEGIN

  r64 := 0;
  i   := 0;
  n   := 0;
  k   := 0;

  REPEAT
    IF n < 21 THEN BEGIN
      d[n] := ORD(ch)-$30;
      INC(n)
    END ELSE BEGIN
      MCB.Mark('too many digits');
      n := 0;
    END;
    Read(ch)
  UNTIL (ch < '0') OR (ch > '9') AND (ch < 'A') OR (ch > 'F');

  IF (ch = 'H') OR (ch = 'R') OR (ch = 'X') THEN BEGIN

    REPEAT
      h := d[i];
      IF h >= 10 THEN h := h-7;
      k := k*$10 + h;
      INC(i);
    UNTIL i = n;

    IF ch = 'X' THEN BEGIN
      sym := lx_char;
      IF k < $100 THEN BEGIN
        r64 := k;
      END ELSE BEGIN
        MCB.Mark('illegal value');
        r64 := 0;
      END;
    END ELSE IF ch = 'R' THEN BEGIN
      sym := lx_real;
      IF (x >= MCB.minreal) AND (x <= MCB.maxreal) THEN BEGIN
        r64 := x;
      END ELSE IF (x >= MCB.minlreal) AND (x <= MCB.maxlreal) THEN BEGIN
        sym := lx_lreal;
        r64 := x;
      END ELSE BEGIN
        r64  := 0;
        MCB.Mark('bad real');
      END;
    END ELSE BEGIN
      sym := lx_int;
      IF (k >= MCB.minint) AND (k <= MCB.maxint) THEN BEGIN
        r64 := k
      END ELSE IF (k >= MCB.minhuge) AND (k <= MCB.maxhuge) THEN BEGIN
        sym := lx_huge;
        r64 := k;
      END ELSE BEGIN
        r64 := 0;
        MCB.Mark('int too large');
      END;
    END;

    Read(ch);

  END ELSE IF ch = '.' THEN BEGIN

    Read(ch);

    IF ch = '.' THEN BEGIN

      ch := #$7F;
      REPEAT
        IF d[i] < 10 THEN BEGIN
          IF k <= RTL._DIV(MCB.maxhuge-d[i],10) THEN
            k := k *10 + d[i]
          ELSE BEGIN
            MCB.Mark('int too large');
            k := 0;
          END;
        END ELSE
          MCB.Mark('bad integer');
        INC(i);
      UNTIL i = n;

      sym := lx_int;
      IF (k >= MCB.minint) AND (k <= MCB.maxint) THEN BEGIN
        r64 := k
      END ELSE IF (k >= MCB.minhuge) AND (k <= MCB.maxhuge) THEN BEGIN
        sym := lx_huge;
        r64 := k;
      END ELSE BEGIN
        r64 := 0;
        MCB.Mark('int too large');
      END;

    END ELSE BEGIN

      x := 0.0;
      e := 0;

      REPEAT
        x := x * 10.0 + d[i]; //RTL._FLT(d[i]);
        INC(i)
      UNTIL i = n;

      WHILE (ch >= '0') AND (ch <= '9') DO BEGIN
        x := x * 10.0 + (ORD(ch) - $30);//RTL._FLT(ORD(ch) - $30);
        DEC(e);
        Read(ch);
      END;

      IF (ch = 'E') OR (ch = 'D') THEN BEGIN

        Read(ch);

        s := 0;
        IF ch = '-' THEN BEGIN
          negE := TRUE;
          Read(ch);
        END ELSE BEGIN
          negE := FALSE;
          IF ch = '+' THEN Read(ch);
        END;

        IF (ch >= '0') AND (ch <= '9') THEN BEGIN
          REPEAT
            s := s*10 + ORD(ch)-$30;
            Read(ch);
          UNTIL (ch < '0') OR (ch >'9');
          IF negE THEN e := e-s ELSE e := e+s;
        END ELSE
          MCB.Mark('digit?');
      END ;

      IF e < 0 THEN BEGIN
        IF e >= MCB.minexp - 1 THEN
          x := x / Ten(-e)
        ELSE BEGIN
          x := 0.0;
          MCB.Mark('real too small')
        END;
      END ELSE IF e > 0 THEN BEGIN
        IF e <= MCB.maxexp THEN
          x := Ten(e) * x
        ELSE BEGIN
          x := 0.0;
          MCB.Mark('real too large')
        END;
      END;

      sym := lx_real;
      IF (x >= MCB.minreal) AND (x <= MCB.maxreal) THEN BEGIN
        r64 := x;
      END ELSE IF (x >= MCB.minlreal) AND (x <= MCB.maxlreal) THEN BEGIN
        sym := lx_lreal;
        r64 := x;
      END ELSE BEGIN
        r64  := 0;
        MCB.Mark('bad real');
      END;

    END;

  END ELSE BEGIN

    REPEAT
      IF d[i] < 10 THEN BEGIN
        IF k <= RTL._DIV(MCB.maxhuge-d[i],10) THEN BEGIN
          k := k*10 + d[i];
        END ELSE BEGIN
          MCB.Mark('int too large');
          k := 0;
        END;
      END ELSE
        MCB.Mark('bad integer');
      INC(i);
    UNTIL i = n;

    sym := lx_int;
    IF (k >= MCB.minint) AND (k <= MCB.maxint) THEN BEGIN
      r64 := k
    END ELSE IF (k >= MCB.minhuge) AND (k <= MCB.maxhuge) THEN BEGIN
      sym := lx_huge;
      r64 := k;
    END ELSE BEGIN
      r64 := 0;
      MCB.Mark('int too large');
    END;

  END;

END;

PROCEDURE Get(VAR sym: INTEGER);
BEGIN

  WHILE NOT EOF AND (ch = ' ') DO Read(ch);
  MCB.prevline := line;
  MCB.prevcol  := col;

  REPEAT
    WHILE NOT EOF AND (ch <= ' ') DO Read(ch);

    IF (MCB.errcount > 0) THEN BEGIN
      MCB.prevline := line;
      MCB.prevcol  := col;
    END;

    IF EOF THEN
      sym := lx_eot
    ELSE IF ch <= #$7F THEN BEGIN
      CASE ch OF
        '0'..'9'         : BEGIN Number(sym); END;
        'a'..'z','A'..'Z': BEGIN Identifier(sym); END;
        '"' : BEGIN TextString; sym := lx_string; END;
        '$' : BEGIN HexString;  sym := lx_string; END;
        '(' : BEGIN
                Read(ch);
                IF ch = '*' THEN BEGIN
                  sym := lx_null;
                  Comment;
                END ELSE
                  sym := lx_lparen;
              END;
        '.' : BEGIN
                Read(ch);
                IF ch = '.' THEN BEGIN
                  Read(ch);
                  sym := lx_upto;
                END ELSE
                  sym := lx_period;
              END;
        ':' : BEGIN
                Read(ch);
                IF ch = '=' THEN BEGIN
                  Read(ch);
                  sym := lx_becomes;
                END ELSE
                  sym := lx_colon;
              END;
        '<' : BEGIN
                Read(ch);
                IF ch = '=' THEN BEGIN
                  Read(ch);
                  sym := lx_leq;
                END ELSE
                  sym := lx_lss;
              END;
        '>' : BEGIN
                Read(ch);
                IF ch = '=' THEN BEGIN
                  Read(ch);
                  sym := lx_geq;
                END ELSE
                  sym := lx_gtr;
              END;
        '!','%','''','?',
        '@','\','_','`':
              BEGIN Read(ch); sym := lx_null;      END;
        '=' : BEGIN Read(ch); sym := lx_eql;       END;
        ';' : BEGIN Read(ch); sym := lx_semicolon; END;
        '/' : BEGIN Read(ch); sym := lx_rdiv;      END;
        '#' : BEGIN Read(ch); sym := lx_neq;       END;
        '&' : BEGIN Read(ch); sym := lx_and;       END;
        ')' : BEGIN Read(ch); sym := lx_rparen;    END;
        '*' : BEGIN Read(ch); sym := lx_times;     END;
        '+' : BEGIN Read(ch); sym := lx_plus;      END;
        ',' : BEGIN Read(ch); sym := lx_comma;     END;
        '-' : BEGIN Read(ch); sym := lx_minus;     END;
        '[' : BEGIN Read(ch); sym := lx_lbrak;     END;
        ']' : BEGIN Read(ch); sym := lx_rbrak;     END;
        '^' : BEGIN Read(ch); sym := lx_arrow;     END;
        '{' : BEGIN Read(ch); sym := lx_lbrace;    END;
        '|' : BEGIN Read(ch); sym := lx_bar;       END;
        '}' : BEGIN Read(ch); sym := lx_rbrace;    END;
        '~' : BEGIN Read(ch); sym := lx_not;       END;
        #$7F: BEGIN Read(ch); sym := lx_upto;      END;
      END;
    END ELSE BEGIN
      IF ch IN  [''..'','',''..'',''] THEN BEGIN
        sym := lx_utf;
        WHILE (ch IN  [''..'','',''..'','']) AND NOT EOF DO Read(ch);
      END ELSE BEGIN
        Read(ch);
        sym := lx_null;
      END;
    END;
  UNTIL sym <> lx_null;
END;

PROCEDURE EnterKW(sym: INTEGER; name: ARRAY OF CHAR);
BEGIN
  keyTab[k].id := COPY(name,0,length(name));
  keyTab[k].sym := sym;
  INC(k)
END;

BEGIN
  k        := 0;
  KWX[0]   := 0;
  KWX[1]   := 0;

  EnterKW(lx_if, 'IF');
  EnterKW(lx_do, 'DO');
  EnterKW(lx_of, 'OF');
  EnterKW(lx_or, 'OR');
  EnterKW(lx_to, 'TO');
  EnterKW(lx_in, 'IN');
  EnterKW(lx_is, 'IS');
  EnterKW(lx_by, 'BY');
  KWX[2] := k;
  EnterKW(lx_end, 'END');
  EnterKW(lx_nil, 'NIL');
  EnterKW(lx_var, 'VAR');
  EnterKW(lx_div, 'DIV');
  EnterKW(lx_mod, 'MOD');
  EnterKW(lx_for, 'FOR');
  KWX[3] := k;
  EnterKW(lx_else, 'ELSE');
  EnterKW(lx_then, 'THEN');
  EnterKW(lx_true, 'TRUE');
  EnterKW(lx_type, 'TYPE');
  EnterKW(lx_case, 'CASE');
  KWX[4] := k;
  EnterKW(lx_elsif, 'ELSIF');
  EnterKW(lx_false, 'FALSE');
  EnterKW(lx_array, 'ARRAY');
  EnterKW(lx_begin, 'BEGIN');
  EnterKW(lx_const, 'CONST');
  EnterKW(lx_until, 'UNTIL');
  EnterKW(lx_while, 'WHILE');
  KWX[5] := k;
  EnterKW(lx_record, 'RECORD');
  EnterKW(lx_repeat, 'REPEAT');
  EnterKW(lx_return, 'RETURN');
  EnterKW(lx_import, 'IMPORT');
  EnterKW(lx_module, 'MODULE');
  KWX[6] := k;
  EnterKW(lx_pointer, 'POINTER');
  KWX[7] := k;
  KWX[8] := k;
  EnterKW(lx_procedure, 'PROCEDURE');
  KWX[9] := k;
  EnterKW(lx_definition, 'DEFINITION');
  KWX[10] := k;

  scandata := NIL;
  namelist := NIL;

END.

