unit SynOberon;

interface

uses
  SysUtils, LCLProc, Dialogs, Classes, Graphics, SynEdit,
  SynEditHighlighter, SynEditHighlighterFoldBase, SynGutterCodeFolding;

type

  TRangeState = (rsComment,rsImport,rsSpecial,rsExport);

  TtkTokenKind = (tkAsm, tkComment, tkHex, tkIdentifier, tkKey, tkNull, tkNumber,
    tkSpace, tkAnsiString, tkString, tkSymbol, tkSpecial, tkUnknown, tkSet,
    tkTypes, tkSystem, tkSysmod, tkDirective, tkExport);

  TSynOberon = class(TSynCustomFoldHighlighter)
  private
    fLineStr             : string;
    fLine                : PChar;
    fLineLen             : integer;
    Run                  : LongInt;
    fStringLen           : Integer;
    fTokenPos            : Integer;
    fTokenID             : TtkTokenKind;
    fLineNumber          : integer;
    fRange               : int64;
    fStringAttri         : TSynHighlighterAttributes;
    fExportAttri         : TSynHighlighterAttributes;
    fNumberAttri         : TSynHighlighterAttributes;
    fKeyAttri            : TSynHighlighterAttributes;
    fSymbolAttri         : TSynHighlighterAttributes;
    fSetAttri            : TSynHighlighterAttributes;
    fCommentAttri        : TSynHighlighterAttributes;
    fHexAttri            : TSynHighlighterAttributes;
    fDirectiveAttri      : TSynHighlighterAttributes;
    fIdentifierAttri     : TSynHighlighterAttributes;
    fSpaceAttri          : TSynHighlighterAttributes;
    fSpecialAttri        : TSynHighlighterAttributes;

    fTypAttri            : TSynHighlighterAttributes;
    fSysAttri            : TSynHighlighterAttributes;
    fModAttri            : TSynHighlighterAttributes;

    procedure   NullProc;

  public
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;
    function    GetEol: Boolean; override;
    function    GetRange: Pointer; override;
    function    GetToken: string; override;
    procedure   GetTokenEx(out TokenStart: PChar; out TokenLength: integer); override;
    function    GetTokenAttribute: TSynHighlighterAttributes; override;
    function    GetTokenID: TtkTokenKind;
    function    GetTokenKind: integer; override;
    function    GetTokenPos: Integer; override;
    function    GetDefaultAttribute(Index: integer): TSynHighlighterAttributes; override;

    procedure   Next; override;

    procedure   ResetRange; override;
    procedure   SetLine({$IFDEF FPC}const {$ENDIF}NewValue: string; LineNumber: Integer); override;
    procedure   SetRange(Value: Pointer); override;
    procedure   IncRange(Value: TRangeState);
    procedure   DecRange(Value: TRangeState);
    procedure   NullRange(Value: TRangeState);
    function    ValRange(Value: TRangeState): integer;

  published
    property SetAttri         : TSynHighlighterAttributes read fSetAttri        write fSetAttri;
    property CommentAttri     : TSynHighlighterAttributes read fCommentAttri    write fCommentAttri;
    property HexAttri         : TSynHighlighterAttributes read fHexAttri        write fHexAttri;
    property DirectiveAttri   : TSynHighlighterAttributes read fDirectiveAttri  write fDirectiveAttri;
    property IdentifierAttri  : TSynHighlighterAttributes read fIdentifierAttri write fIdentifierAttri;
    property KeyAttri         : TSynHighlighterAttributes read fKeyAttri        write fKeyAttri;
    property NumberAttri      : TSynHighlighterAttributes read fNumberAttri     write fNumberAttri;
    property SpaceAttri       : TSynHighlighterAttributes read fSpaceAttri      write fSpaceAttri;
    property ExportAttri      : TSynHighlighterAttributes read fExportAttri     write fExportAttri;
    property StringAttri      : TSynHighlighterAttributes read fStringAttri     write fStringAttri;
    property SymbolAttri      : TSynHighlighterAttributes read fSymbolAttri     write fSymbolAttri;
    property SpecialAttri     : TSynHighlighterAttributes read fSpecialAttri    write fSpecialAttri;
    property TypAttri         : TSynHighlighterAttributes read fTypAttri        write fTypAttri;
    property SysAttri         : TSynHighlighterAttributes read fSysAttri        write fSysAttri;
    property ModAttri         : TSynHighlighterAttributes read fModAttri        write fModAttri;
  end;

implementation

var
  KeywordsList: TStringList;

const

  TokenKey : array [0..48] of string[10]=(
    'IF','DO','OF','OR','TO','IN','IS','BY',
    'END','NIL','VAR','DIV','MOD','FOR',
    'INC','DEC','INCL','EXCL','NEW','ASSERT',
    'PACK','UNPK','DISPOSE','ABS','ODD','LEN',
    'LSL','ASR','ROR','FLOOR','FLT','ORD','CHR',
    'ELSE','THEN','TYPE','CASE',
    'ELSIF','ARRAY','BEGIN','CONST','UNTIL','WHILE',
    'RECORD','REPEAT','MODULE','POINTER','ADDRESS','PROCEDURE'
  );

  TypesKey : array [0..13] of string[10]=(
    'INTEGER','INT8','INT16','INT32','BYTE','CARD16','CARD32',
    'BOOLEAN','SET','CHAR','CHAR8','REAL','LONGINT','LONGREAL'
  );

  SystemKey : array [0..4] of string[10]=(
    'TRUE','FALSE','IMPORT','RETURN','EXIT'
  );

  SystemMod : array [0..7] of string[15]=(
    'SYSTEM','SYSTEM.GET','SYSTEM.PUT','SYSTEM.COPY',
    'SYSTEM.ADR','SYSTEM.SIZE','SYSTEM.BIT','SYSTEM.VAL'
  );

constructor TSynOberon.Create(AOwner: TComponent);
var
  i: char;
  n: integer;
begin
  inherited Create(AOwner);

  if KeywordsList = nil then begin
    KeywordsList := TStringList.Create;
    for n:=0 to High(TokenKey)      do KeywordsList.AddObject(TokenKey[n],     TObject(tkKey));
    for n:=0 to High(TypesKey)      do KeywordsList.AddObject(TypesKey[n],     TObject(tkTypes));
    for n:=0 to High(SystemKey)     do KeywordsList.AddObject(SystemKey[n],    TObject(tkSystem));
    for n:=0 to High(SystemMod)     do KeywordsList.AddObject(SystemMod[n],    TObject(tkSysmod));
    KeywordsList.Sort;
    KeywordsList.Sorted := FALSE;
  end;

  fSetAttri := TSynHighlighterAttributes.Create('Set', 'Set');
  fSetAttri.Foreground := $00996600;
  fSetAttri.Style     := [fsItalic];
  AddAttribute(fSetAttri);

  fCommentAttri := TSynHighlighterAttributes.Create('Comment', 'Comment');
  fCommentAttri.Style:= [fsItalic];
  fCommentAttri.Foreground := $4e8b60;

  AddAttribute(fCommentAttri);

  fHexAttri := TSynHighlighterAttributes.Create('Hex', 'Hex');
  fHexAttri.Style:= [];
  fHexAttri.Foreground := $00008800;
  AddAttribute(fHexAttri);

  fDirectiveAttri := TSynHighlighterAttributes.Create('Directive', 'Directive');
  fDirectiveAttri.Style:= [];
  fDirectiveAttri.Foreground := $00888800;
  AddAttribute(fDirectiveAttri);

  fKeyAttri := TSynHighlighterAttributes.Create('Reserved', 'Reserved');
  fKeyAttri.Style     := [];
  fKeyAttri.Foreground := $d69c56;
  AddAttribute(fKeyAttri);

  fTypAttri           := TSynHighlighterAttributes.Create('Type','Type');
  fTypAttri.Style     := [];
  fTypAttri.Foreground := $b0c94e;
  AddAttribute(fTypAttri);

  fSysAttri           := TSynHighlighterAttributes.Create('System','System');
  fSysAttri.Style     := [];
  fSysAttri.Foreground := $c086c5;
  AddAttribute(fSysAttri);

  fModAttri           := TSynHighlighterAttributes.Create('Module','Module');
  fModAttri.Style     := [];
  fModAttri.Foreground := $5555ff;
  AddAttribute(fModAttri);

  fNumberAttri := TSynHighlighterAttributes.Create('Number', 'Number');
  fModAttri.Style     := [fsBold];
  fNumberAttri.Foreground := $fedc9c;
  AddAttribute(fNumberAttri);

  fSpaceAttri := TSynHighlighterAttributes.Create('Space', 'Space');
  AddAttribute(fSpaceAttri);

  fStringAttri := TSynHighlighterAttributes.Create('String', 'String');
  fStringAttri.Foreground:=$7891ce;
  fStringAttri.Style:= [];
  AddAttribute(fStringAttri);

  fExportAttri := TSynHighlighterAttributes.Create('Export', 'Export');
  fExportAttri.Foreground:=$BBFFBB;
  fExportAttri.Style:= [];
  AddAttribute(fExportAttri);

  fSymbolAttri := TSynHighlighterAttributes.Create('Symbol', 'Symbol');
  fSymbolAttri.Foreground := $fedc9c;
  AddAttribute(fSymbolAttri);

  fSpecialAttri := TSynHighlighterAttributes.Create('Special', 'Special');
  fSpecialAttri.Style     := [fsBold];
  fSpecialAttri.Foreground := $89b6ce;
  AddAttribute(fSpecialAttri);

  fRange := 0;

end;

function TSynOberon.GetTokenAttribute: TSynHighlighterAttributes;
begin
  case GetTokenID of
    tkSet         : Result := fSetAttri;
    tkComment     : Result := fCommentAttri;
    tkHex         : Result := fHexAttri;
    tkDirective   : Result := fDirectiveAttri;
    tkIdentifier  : Result := fIdentifierAttri;
    tkKey         : Result := fKeyAttri;
    tkTypes       : Result := fTypAttri;
    tkSystem      : Result := fSysAttri;
    tkSysmod      : Result := fModAttri;
    tkNumber      : Result := fNumberAttri;
    tkSpace       : Result := fSpaceAttri;
    tkExport      : Result := fExportAttri;
    tkString      : Result := fStringAttri;
    tkSymbol      : Result := fSymbolAttri;
    tkSpecial     : Result := fSpecialAttri;
  else
    Result := nil;
  end;
end;

function TSynOberon.GetDefaultAttribute(Index: integer): TSynHighlighterAttributes;
begin
  if index>=0 then Result:=nil;
end;

destructor TSynOberon.Destroy;
begin
  inherited Destroy;
end;

procedure TSynOberon.SetLine(const NewValue: string; LineNumber:Integer);
begin
  fLineStr := NewValue;
  fLineLen:=length(fLineStr);
  fLine:=PChar(Pointer(fLineStr));
  fLineNumber := LineNumber;
  Run := 0;
  Inherited SetLine(NewValue,LineNumber);
  Next;
end; { SetLine }

function TSynOberon.GetEol: Boolean;
begin
  Result := (fTokenID = tkNull) and (Run >= fLineLen);
end;

function TSynOberon.GetToken: string;
var
  Len: LongInt;
begin
  Len := Run - fTokenPos;
  SetLength(Result,Len);
  if Len>0 then System.Move(fLine[fTokenPos],Result[1],Len);
end;

procedure TSynOberon.GetTokenEx(out TokenStart: PChar; out TokenLength: integer);
begin
  TokenLength:=Run-fTokenPos;
  if TokenLength>0 then begin
    TokenStart:=@fLine[fTokenPos];
  end else begin
    TokenStart:=nil;
  end;
end;

function TSynOberon.GetTokenKind: integer;
begin
  Result := Ord(GetTokenID);
end;

function TSynOberon.GetTokenID: TtkTokenKind;
begin
  Result := fTokenId;
end;

function TSynOberon.GetTokenPos: Integer;
begin
  Result := fTokenPos;
end;

function TSynOberon.GetRange: Pointer;
begin
  CodeFoldRange.RangeType := pointer(integer(fRange));
  Result := inherited GetRange;
end;

procedure TSynOberon.SetRange(Value: Pointer);
begin
  inherited SetRange(Value);
  fRange := integer(CodeFoldRange.RangeType);
end;

procedure TSynOberon.ResetRange;
begin
  fRange := 0;
  inherited ResetRange;
end;

procedure TSynOberon.IncRange(Value: TRangeState);
begin
  fRange := fRange + (1 shl (ord(Value)*8));
end;

procedure TSynOberon.DecRange(Value: TRangeState);
begin
  if ValRange(Value) > 0 then fRange := fRange - (1 shl (ord(Value)*8));
end;

procedure TSynOberon.NullRange(Value: TRangeState);
begin
  fRange := fRange - (ValRange(Value) shl (ord(Value)*8));
end;

function TSynOberon.ValRange(Value: TRangeState): integer;
begin
  Result := (fRange shr (ord(Value)*8)) mod $100;
end;

(*########################################*)

procedure TSynOberon.NullProc;
begin
  fTokenID := tkNull;
  if Run<fLineLen then inc(Run);
end;

procedure TSynOberon.Next;
var
  found : boolean;
  search: boolean;
  c     : AnsiChar;
  start : integer;
  word  : string;
  i,n   : integer;
begin
  fTokenPos := Run;

  if Run>=fLineLen then begin
    NullProc;
    exit;
  end;

  if ValRange(rsComment) = 0 then
    fTokenID := tkUnknown
  else
    fTokenID := tkComment;

  case fLine[Run] of

    '(':
      begin
        fTokenID := tkComment;
        if ValRange(rsComment) = 0 then fTokenID := tkSymbol;
        Inc(Run);
        if fLine[Run]='*' then begin
          fTokenID := tkComment;
          IncRange(rsComment);
          Inc(Run);
        end;
      end;

    '*':
      begin
        fTokenID := tkComment;
        if ValRange(rsComment) = 0 then fTokenID := tkSymbol;
        Inc(Run);
        if fLine[Run]=')' then begin
          fTokenID := tkComment;
          if (ValRange(rsComment) > 0) then DecRange(rsComment);
          Inc(Run);
          if ValRange(rsImport) > 0 then NullRange(rsImport);
        end;
      end;

    ';':
      begin
        fTokenID := tkSymbol;
        if ValRange(rsImport) > 0 then NullRange(rsImport);
        Inc(Run);
      end;

    #128..#255:
      begin
        Inc(Run);
        while (Run < fLineLen) and (fLine[Run] in [#128..#255]) do inc(Run);
      end;

  else
    if ValRange(rsComment) = 0 then begin

      case fLine[Run] of

        '''','"'                      :
                                      begin
                                        fTokenID := tkString;
                                        c := fLine[Run];
                                        inc(Run);
                                        found := false;
                                        while not found and (Run < fLineLen) do begin
                                          found := fLine[Run] = c;
                                          inc(Run);
                                        end;
                                      end;

        '0'..'9'                      :
                                      begin
                                        fTokenID := tkNumber;
                                        inc(Run);
                                        while fLine[Run] in ['0'..'9'] do inc(Run);
                                      end;

        'a'..'z','A'..'Z','_'       :
                                      begin
                                        start := Run;
                                        inc(Run);
                                        fTokenID := tkIdentifier;
                                        while fLine[Run] in ['0'..'9','a'..'z','A'..'Z','_','.'] do inc(Run);
                                        word := copy(fLine,start+1,Run-start);

                                        n := 0;
                                        search := false;
                                        while not search and (n < KeywordsList.Count) do begin
                                          if (word<>'') and (word[1] = KeywordsList[n][1])
                                          and (word = KeywordsList[n]) then begin
                                            fTokenID := TtkTokenKind(PtrUInt(KeywordsList.Objects[n]));
                                            search   := true;
                                            case Word[1] of
                                              'B': if  Word='BEGIN'     then StartCodeFoldBlock(nil);
                                              'C': if  Word='CASE'      then StartCodeFoldBlock(nil);
                                              'F': if  Word='FOR'       then StartCodeFoldBlock(nil);
                                              'I': if  Word='IF'        then StartCodeFoldBlock(nil);
                                              'L': if  Word='LOOP'      then StartCodeFoldBlock(nil);
                                              'R': if  Word='RECORD'    then StartCodeFoldBlock(nil);
                                              'W': if  Word='WHILE'     then StartCodeFoldBlock(nil);
                                              'E': if  Word='END'       then EndCodeFoldBlock();
                                            end;
                                            case Word[1] of
                                              'M': if (Word='MODULE')    and (ValRange(rsExport) = 0) then IncRange(rsExport);
                                              'P': if (Word='PROCEDURE') and (ValRange(rsExport) = 0) then IncRange(rsExport);
                                              'B': if  Word='BEGIN'      then NullRange(rsExport);
                                              'I': if  Word='IMPORT'     then IncRange(rsImport);
                                            end;
                                          end else begin
                                            if ValRange(rsImport) > 0 then fTokenID := tkSpecial;

                                            if ValRange(rsExport) > 0 then begin
                                              start := Run;
                                              while (start <= fLineLen) and (fLine[start] in [' ',#9]) do inc(start);
                                              if fLine[start] = '*' then begin
                                                fTokenID := tkExport;
                                                Run := start+1;
                                              end;
                                            end;

                                          end;
                                          inc(n);
                                        end;
                                      end;

        '{','}', '!', '%', '&','|',
        '('..'/', ':'..'@','#',
        '['..'^', '`', '~','$'      :
                                     begin
                                       fTokenID := tkSymbol;
                                       inc(Run);
                                     end;
      else
        Inc(Run);
      end;

    end else begin
      Inc(Run);
    end;

  end;

end;

initialization
  RegisterPlaceableHighlighter(TSynOberon);

finalization
  FreeAndNil(KeywordsList);

end.


