UNIT Unit1;

{$mode objfpc}{$H+}

INTERFACE

USES
  Classes, SysUtils, SynEdit, Forms, Controls, Graphics, Clipbrd,
  ExtCtrls, StdCtrls, Menus, SynOberon,
  SynEditKeyCmds, SynEditMarkupWordGroup,
  Windows, RTL, MCB, MCS, MCE;

TYPE

  { TForm1 }

  TForm1 = CLASS(TForm)
    Button1: TButton;
    Button2: TButton;
    MenuItem1: TMenuItem;
    MenuItem2: TMenuItem;
    MenuItem3: TMenuItem;
    MenuItem4: TMenuItem;
    MenuItem7: TMenuItem;
    MenuItem8: TMenuItem;
    MenuItem9: TMenuItem;
    Panel1: TPanel;
    Panel2: TPanel;
    Panel3: TPanel;
    Panel4: TPanel;
    PopupMenu1: TPopupMenu;
    Splitter1: TSplitter;
    Splitter2: TSplitter;
    SynEdit1: TSynEdit;
    PROCEDURE Button1Click(Sender: TObject);
    PROCEDURE Button2Click(Sender: TObject);
    PROCEDURE FormActivate(Sender: TObject);
    PROCEDURE FormCreate(Sender: TObject);
    PROCEDURE FormDestroy(Sender: TObject);
    PROCEDURE MenuItem1Click(Sender: TObject);
    PROCEDURE MenuItem2Click(Sender: TObject);
    PROCEDURE MenuItem3Click(Sender: TObject);
    PROCEDURE MenuItem4Click(Sender: TObject);
    PROCEDURE MenuItem7Click(Sender: TObject);
    PROCEDURE MenuItem9Click(Sender: TObject);
    PROCEDURE Panel3Resize(Sender: TObject);
    PROCEDURE Panel4Paint(Sender: TObject);
  protected
    PROCEDURE UpdateActions; override;
  private
    node   : MCB.Item;
    console: INTEGER;
    PROCEDURE SetConsole;
    PROCEDURE DrawItem(VAR i: MCB.Item; w,h,l: INTEGER; VAR ofs: INTEGER);
  public

  END;

VAR
  Form1: TForm1;

IMPLEMENTATION

{$R *.lfm}

{ TForm1 }

PROCEDURE MarkProc(msg: ARRAY OF CHAR);
BEGIN
  IF MCB.errcount = 0 THEN BEGIN
    WriteLn('[',MCB.prevline,':',MCB.prevcol,']',COPY(msg,0,LENGTH(msg)));
    Form1.SynEdit1.CaretXY := Classes.POINT(MCB.prevcol,MCB.prevline);
    Form1.SynEdit1.SetFocus();
    MCB.errpos := MCB.prevcol;
  END ELSE IF (MCB.errpos <> MCB.prevcol) AND (MCB.errcount < 25) THEN BEGIN
    WriteLn('[',MCB.prevline,':',MCB.prevcol,']',COPY(msg,0,LENGTH(msg)));
  END;
  INC(MCB.errcount);
  MCB.errpos := MCB.prevcol;
END;

PROCEDURE TForm1.FormCreate(Sender: TObject);
VAR
  hlg: TSynOberon;
BEGIN
  MCB.Mark := @MarkProc;
  width  := Screen.Width  - Screen.Width  DIV 6;
  height := Screen.Height - Screen.Height DIV 4;
  hlg    := TSynOberon.Create(self);
  SynEdit1.Highlighter := hlg;
  TSynEditMarkupWordGroup(SynEdit1.MarkupByClass[TSynEditMarkupWordGroup]).Enabled := false;
END;

PROCEDURE TForm1.FormActivate(Sender: TObject);
BEGIN
  SetConsole;
  SynEdit1.SetFocus();
END;

PROCEDURE TForm1.FormDestroy(Sender: TObject);
BEGIN
  IF console > 0 THEN FreeConsole;
END;

PROCEDURE TForm1.SetConsole;
BEGIN
  AllocConsole;
  console := GetConsoleWindow();
  IF console > 0 THEN BEGIN
    ShowWindow(console,SW_HIDE);
    ReWrite(Output);
    TTextRec(Output).Handle := GetStdHandle(STD_OUTPUT_HANDLE);
    Windows.SetParent(console, Panel3.Handle);
    SetWindowLong(console, GWL_STYLE, GetWindowLong(console, GWL_STYLE) AND NOT WS_HSCROLL AND NOT WS_VSCROLL AND NOT WS_THICKFRAME AND NOT WS_DLGFRAME AND NOT WS_BORDER);
    SetWindowPos(console, 0, 0,0, Panel3.width - 90, Panel3.height, SWP_NOACTIVATE);
    ShowWindow(console,SW_SHOW);
  END;
END;

PROCEDURE TForm1.Panel3Resize(Sender: TObject);
BEGIN
  SetWindowPos(console, 0, 0,0, Panel3.width - 90, Panel3.height, SWP_NOACTIVATE);
END;

PROCEDURE TForm1.MenuItem1Click(Sender: TObject);
BEGIN
  SynEdit1.CutToClipboard;
END;

PROCEDURE TForm1.MenuItem2Click(Sender: TObject);
BEGIN
  SynEdit1.CopyToClipboard;
END;

PROCEDURE TForm1.MenuItem3Click(Sender: TObject);
BEGIN
  SynEdit1.PasteFromClipboard;
END;

PROCEDURE TForm1.MenuItem4Click(Sender: TObject);
BEGIN
  SynEdit1.ClearSelection;
END;

PROCEDURE TForm1.MenuItem7Click(Sender: TObject);
BEGIN
  SynEdit1.Undo;
END;

PROCEDURE TForm1.MenuItem9Click(Sender: TObject);
BEGIN
  SynEdit1.Redo;
END;

PROCEDURE TForm1.UpdateActions;
VAR
  found: BOOLEAN;
BEGIN
  found := SynEdit1.SelAvail;
  MenuItem1.Enabled := found;
  MenuItem2.Enabled := found;
  MenuItem3.Enabled := Clipboard.HasFormat(CF_TEXT);
  MenuItem4.Enabled := found;
  MenuItem7.Enabled := SynEdit1.CanUndo;
  MenuItem9.Enabled := SynEdit1.CanRedo;
END;

PROCEDURE TForm1.DrawItem(VAR i: MCB.Item; w,h,l: INTEGER; VAR ofs: INTEGER);
VAR
  tw,th: INTEGER;
  s    : STRING;
BEGIN
  INC(l);
  IF i.a <> NIL THEN DrawItem(i.a,w,h,l,ofs);
  i.key := ofs; INC(ofs);
  IF i.b <> NIL THEN DrawItem(i.b,w,h,l,ofs);

  Panel4.canvas.pen.width   := 1;
  Panel4.canvas.pen.color   := clGray;
  IF i.a <> NIL THEN Panel4.canvas.line(w+w+i.key*w-5,l*h+5,w+w+i.a.key*w+6,l*h+h-9);
  IF i.b <> NIL THEN Panel4.canvas.line(w+w+i.key*w+5,l*h+5,w+w+i.b.key*w-6,l*h+h-9);

  CASE i.op OF
    MCS.lx_times  : s := '*';
    MCS.lx_plus   : s := '+';
    MCS.lx_minus  : s := '-';
    MCS.lx_rdiv   : s := '/';
    MCS.lx_div    : s := 'DIV';
    MCS.lx_mod    : s := 'MOD';
    MCS.lx_and    : s := '&';
    MCS.lx_or     : s := 'OR';
  ELSE
    s := i.val.ToString;
  END;

  Panel4.canvas.pen.width   := 1;
  tw := Panel4.canvas.textwidth(s) DIV 2;
  th := Panel4.canvas.textheight('X') DIV 2;
  Panel4.canvas.brush.color := $222222;
  Panel4.canvas.brush.style := bsSolid;
  Panel4.canvas.roundrect(w+w+i.key*w-tw-8,l*h-th-3,w+w+i.key*w+tw+8,l*h+th+4,21,21);
  Panel4.canvas.brush.style := bsClear;
  Panel4.canvas.textout(w+w+i.key*w-tw,l*h-th,s);
END;

PROCEDURE TForm1.Panel4Paint(Sender: TObject);
VAR
  k,wn,hn: INTEGER;

  FUNCTION CalcWH(i: MCB.Item; VAR w: INTEGER): INTEGER;
  VAR
    l,r: INTEGER;
  BEGIN
    IF i<> NIL THEN BEGIN
      l := CalcWH(i.a,w);
      INC(w);
      r := CalcWH(i.b,w);
      IF l > r THEN
        RESULT := l + 1
      ELSE
        RESULT := r + 1;
    END ELSE
      RESULT := 0;
  END;

BEGIN
  IF node <> NIL THEN BEGIN
    k := 0; wn := 0;
    hn := CalcWH(node,wn);
    DrawItem(node,Panel4.width DIV (wn+3),Panel4.height DIV (hn+1),0,k);
  END;
END;

PROCEDURE TForm1.Button2Click(Sender: TObject);
BEGIN
  SynEdit1.Visible := NOT SynEdit1.Visible;
  Button1.Enabled  := SynEdit1.Visible;
  IF SynEdit1.Visible THEN
    SynEdit1.SetFocus()
  ELSE IF node <> NIL THEN
    InvalidateRect(Panel4.handle,NIL,FALSE);
END;

PROCEDURE TForm1.Button1Click(Sender: TObject);
VAR
  x: MCB.Item;
BEGIN
  MCB.Clear;
  MCS.Init(@SynEdit1.Lines.Text[1]);

  MCS.Get(sym);
  MCE.Expression(x);
  node := NIL;
  IF MCB.errcount = 0 THEN BEGIN
    writeln('OK... ');
    node := x;
    IF node <> NIL THEN InvalidateRect(Panel4.handle,NIL,FALSE);
  END;

  SynEdit1.SetFocus();
END;

END.

